from flask_wtf import Form as WTForm
import wtforms.fields as wtfields
import wtforms.fields.html5 as wtfields5
import wtforms.validators as wtvalidators
from frasco import Markup, render_template, AttrDict, translate, ntranslate, lazy_translate
from frasco.templating import parse_template, jinja_node_to_python
from frasco.declarative import FileLoader
from jinja2 import nodes
import inflection
import os


__all__ = ('Form', 'TemplateForm', 'fields', 'validators', 'FormDefinitionError',
           'NoFormError', 'FormLoader', 'create_from_template')


field_type_map = AttrDict({
    "checkbox": wtfields.BooleanField,
    "decimal": wtfields.DecimalField,
    "date": wtfields.DateField,
    "datetime": wtfields.DateTimeField,
    "float": wtfields.FloatField,
    "int": wtfields.IntegerField,
    "radio": wtfields.RadioField,
    "select": wtfields.SelectField,
    "selectmulti": wtfields.SelectMultipleField,
    "text": wtfields.StringField,
    "textarea": wtfields.TextAreaField,
    "password": wtfields.PasswordField,
    "hidden": wtfields.HiddenField,
    "date5": wtfields5.DateField,
    "datetime5": wtfields5.DateTimeField,
    "datetimelocal": wtfields5.DateTimeLocalField,
    "decimal5": wtfields5.DecimalField,
    "decimalrange": wtfields5.DecimalRangeField,
    "email": wtfields5.EmailField,
    "int5": wtfields5.IntegerField,
    "intrange": wtfields5.IntegerRangeField,
    "search": wtfields5.SearchField,
    "tel": wtfields5.TelField,
    "url": wtfields5.URLField
})


# for easy import
fields = field_type_map
validators = wtvalidators


class FormTranslations(object):
    def gettext(self, string, **kwargs):
        return translate(string, **kwargs)

    def ngettext(self, singular, plural, n):
        return translate(singular) if n == 1 else translate(plural)


class Form(WTForm):
    @property
    def enctype(self):
        for f in self:
            if f.type == "FileField":
                return "multipart/form-data"
        return "application/x-www-form-urlencoded"

    def populate_obj(self, obj, ignore_fields=None):
        for name, field in self._fields.iteritems():
            if not ignore_fields or name not in ignore_fields:
                field.populate_obj(obj, name)

    def _get_translations(self):
        return FormTranslations()


class TemplateForm(Form):
    """Wraps the form object to make it easer to build and access form from
    context variables and templates. Access to undefined attributes will
    return a function which when called will return the field which name
    matches the fist argument. All other arguments are ignored.
    """
    template = None

    def __init__(self, template=None, *args, **kwargs):
        super(TemplateForm, self).__init__(*args, **kwargs)
        if template:
            self.template = template

    def render(self, **kwargs):
        return render_template(self.template, form=self, **kwargs)

    def __call__(self, **kwargs):
        return Markup(self.render(**kwargs))


def _fake_field_init(field):
    def caller(*args, **kwargs):
        return field
    return caller


class TemplateField(object):
    _formfield = True

    def __init__(self, field):
        object.__setattr__(self, "_field", field)

    def __getattr__(self, name):
        if hasattr(self._field, name):
            return getattr(self._field, name)
        if name in field_type_map:
            return _fake_field_init(self._field)
        raise AttributeError

    def __setattr__(self, name, value):
        setattr(self._field, name, value)

    def __call__(self, *args, **kwargs):
        return self._field(*args, **kwargs)


class UnboundTemplateField(TemplateField):
    def bind(self, *args, **kwargs):
        return TemplateField(self._field.bind(*args, **kwargs))


class FormDefinitionError(Exception):
    pass


class NoFormError(Exception):
    pass


class FormLoader(FileLoader):
    def __init__(self, path, base_template_path="", **kwargs):
        super(FormLoader, self).__init__(path, **kwargs)
        self.base_template_path = base_template_path

    def load_file(self, app, pathname, relpath, pypath):
        name, _ = os.path.splitext(os.path.basename(pathname))
        return create_from_template(app, os.path.join(self.base_template_path, relpath), name)


def create_from_template(app, filename, *args, **kwargs):
    node = parse_template(app, filename)
    form_class = create_from_template_node(node, *args, **kwargs)
    form_class.template = filename
    return form_class


def create_from_template_node(template_node, name="F", var_name="form"):
    # The parsing process looks for Call nodes applied on a GetAttr node
    # where the target is the form variable. The form variable name is "form"
    # by default but can be overrided by the "as" option
    fields = {}
    for call in template_node.find_all((nodes.Call,)):
        if not isinstance(call.node, nodes.Getattr) or not isinstance(call.node.node, nodes.Getattr):
            continue
        attrnode = call.node.node
        if not isinstance(attrnode.node, nodes.Name) or not attrnode.node.name == var_name:
            continue
        fname = attrnode.attr
        fields[fname] = [call.node.attr, {}]

        if len(call.args) > 0:
            fields[fname][1]["label"] = jinja_node_to_python(call.args[0])

        for arg in call.kwargs:
            fields[fname][1][arg.key] = jinja_node_to_python(arg.value)

    form_class = type(name, (TemplateForm,), {"name": name})

    for fname, fopts in fields.iteritems():
        ftype, kwargs = fopts
        if ftype not in field_type_map:
            raise FormDefinitionError("Unknown field type '%s'" % ftype)
        validators = []
        if "label" not in kwargs:
            kwargs["label"] = inflection.humanize(fname)
        if kwargs.pop("required", False):
            if ftype == "file":
                validators.append(FileRequired())
            else:
                validators.append(wtvalidators.DataRequired())
        if kwargs.pop("optional", False):
            validators.append(wtvalidators.Optional())
        if "range" in kwargs:
            min, max = kwargs.pop("range")
            validators.append(wtvalidators.NumberRange(min, max))
        if "length" in kwargs:
            min, max = kwargs.pop("length")
            validators.append(wtvalidators.Length(min, max))
        if "validators" in kwargs:
            for v in kwargs.pop("validators"):
                validators.append(getattr(wtvalidators, v)())
        if "placeholder" in kwargs:
            kwargs.setdefault('render_kw', {})["placeholder"] = kwargs.pop("placeholder")
        setattr(form_class, fname, UnboundTemplateField(field_type_map[ftype](validators=validators, **kwargs)))

    return form_class



from setuptools import setup, find_packages


setup(
    name='frasco-forms',
    version='0.3',
    url='http://github.com/frascoweb/frasco-forms',
    license='MIT',
    author='Maxime Bouroumeau-Fuseau',
    author_email='maxime.bouroumeau@gmail.com',
    description="WTForms integration for Frasco",
    packages=find_packages(),
    package_data={
        'frasco_forms': ['*.html']
    },
    zip_safe=False,
    platforms='any',
    install_requires=[
        'frasco',
        'Flask-WTF>=0.14.0',
        'inflection'
    ]
)



from frasco import Feature, action, signal, current_app, request, current_context
from frasco.utils import AttrDict, DictObject
from jinja2 import FileSystemLoader
import os
import inspect
from .form import *


class FormsFeature(Feature):
    """Create and use forms using wtforms.
    """
    name = "forms"
    defaults = {"import_macros": True,
                "csrf_enabled": True}

    form_validation_failed_signal = signal("form_validation_failed")
    form_submitted_signal = signal("form_submitted")
    form_created_from_view_signal = signal("form_created_from_view")

    def init_app(self, app):
        self.forms = {}
        if self.options["import_macros"]:
            macro_file = os.path.join(os.path.dirname(__file__), "macros.html")
            app.jinja_env.macros.register_file(macro_file, "form.html")
            app.jinja_env.macros.alias("form_tag", "form")

    def init_declarative(self, app):
        self.load_forms_from_templates(app)

    def load_forms_from_templates(self, app, folder="forms", prefix="__forms__"):
        path = os.path.join(app.root_path, folder)
        if os.path.exists(path):
            app.jinja_env.loader.add_prefix(prefix, FileSystemLoader(path))
            for form_class in FormLoader(path, prefix).load(app):
                self.register(form_class)

    def register(self, form_class):
        self.forms[form_class.__name__] = form_class
        return form_class

    def __getitem__(self, name):
        return self.forms[name]

    def __setitem__(self, name, form):
        self.forms[name] = form

    def __contains__(self, name):
        return name in self.forms

    @action(default_option="form", methods=("GET", "POST"), as_="form")
    def form(self, obj=None, form=None, name=None, template=None, var_name=None, validate_on_submit=True,\
        exit_on_failure=True, csrf_enabled=None):
        """Loads a form and validates it (unless specified).
        If the form referenced has not been loaded, an attempt to create a form
        object using the information in the template will be made.
        """
        if not form or isinstance(form, str):
            if not name and isinstance(form, str):
                name = form
            if not name:
                name = request.endpoint.rsplit('.', 1)[1] if '.' in request.endpoint else request.endpoint
            if name not in self.forms:
                raise NoFormError("Cannot find form '%s'" % name)
            form = self.forms[name]
        if inspect.isclass(form):
            if isinstance(obj, dict):
                obj = DictObject(obj)
            if csrf_enabled is None:
                csrf_enabled = self.options["csrf_enabled"]
            form = form(obj=obj, meta={'csrf': csrf_enabled})

        current_context.data.form = form
        yield form

        if validate_on_submit and request.method == "POST":
            self.validate(form, exit_on_failure=exit_on_failure)

    @action("create_form_from_view", default_option="name", as_="form_class")
    def create_from_view(self, name=None, template=None, var_name=None):
        if not name:
            name = request.endpoint.rsplit('.', 1)[1] if '.' in request.endpoint else request.endpoint
        return self.forms[name]

    @create_from_view.init_view
    @form.init_view
    def init_form_view(self, view, opts):
        """Checks if the form referenced in the view exists or attempts to
        create it by parsing the template
        """
        name = opts.get("name", opts.get("form"))
        if isinstance(name, Form):
            return

        template = opts.get("template", getattr(view, "template", None))
        if not template:
            if not name:
                raise NoFormError("No form name specified in the form action and no template")
            return

        try:
            as_ = opts.get("var_name", getattr(self.form, "as_", "form"))
            form_class = create_from_template(current_app, template, var_name=as_)
        except NoFormError:
            if not name:
                raise
            return

        if not name:
            name = view.name

        self.forms[name] = form_class
        self.form_created_from_view_signal.send(self, view=view, form_class=form_class)
        return form_class

    @action("create_form_from_template", as_="form_class")
    def create_from_template(self, name, template, var_name="form"):
        form_class = create_from_template(current_app, template, var_name=var_name)
        self.forms[name] = form_class
        return form_class

    @action("validate_form")
    def validate(self, form=None, return_success=False, exit_on_failure=True):
        ctx = current_context
        if not form:
            form = ctx.data.form
        if not form.validate():
            self.form_validation_failed_signal.send(self, form=form)
            if exit_on_failure:
                ctx.exit(trigger_action_group="form_validation_failed")
            ctx.trigger_action_group("form_validation_failed")
        self.form_submitted_signal.send(self, form=form)
        if not return_success:
            ctx.trigger_action_group("form_submitted")
        return True

    @action("form_to_obj")
    def populate_obj(self, obj=None, form=None):
        """Populates an object with the form's data
        """
        if not form:
            form = current_context.data.form
        if obj is None:
            obj = AttrDict()
        form.populate_obj(obj)
        return obj